/****************************************************************************
****************************************************************************/

/*
This is a part of the LiteStep Shell Source code.

Copyright (C) 1997-2001 The LiteStep Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/****************************************************************************
05-01-2002 - Christian Buffin (Jesus_mjjg)
  - //CPU 20020501 to find what has been changed to take less cpu
01-17-2000 - Charles Oliver Nutter (Headius)
  - Added TaskbarWidth (default of screen width), TaskbarPositioning (default
    of false, specifies that TaskbarX and TaskbarY should be used), TaskbarX
	(default of 0), TaskbarY (default of 0)
12-06-2000 - Kevin Schaffer (Maduin)
  - Added TaskbarFont (et al) and TaskbarActiveFont (et al)
11/28/00 - Bobby G. Vinyard (Message)
  - Added multimonitor support
11/24/00 - Richard Lin (Chaku)
  - Fixed the start button text position problem, and the non-reponsive mouse
    clicking when start button is absent
11/22/00 - Richard Lin (Chaku)
  - Added task flashing.
11/15/00 - Kevin Schaffer (Maduin)
  - First version
****************************************************************************/


#include <algorithm>
#include "common.h"
#include <commctrl.h>
#include <shellapi.h>
#include "../lsapi/lsapi.h"
#include "../lsapi/macros.h"
#include "taskbar.h"

#define BUFFER_SIZE		1024

#define WC_TASKBAR		"TaskbarLS"

#define EDGE_BOTTOM		0
#define EDGE_TOP		1

#define TIMER_UPDATE		1
#define TIMER_UPDATE_SCROLL	2

#define MOUSE_NONE		0
#define MOUSE_SELECT	1
#define MOUSE_MENU		2
#define MOUSE_CLOSE		3
#define MOUSE_CLOSE2	4
#define MOUSE_ONTOP		5

#define DIRECTION_RIGHT	0
#define DIRECTION_LEFT	1


// forward declarations
void hideBangCommand(HWND hCaller, const char *args);
void toggleBangCommand(HWND hCaller, const char *args);
void showBangCommand(HWND hCaller, const char *args);

void sizeBangCommand(HWND hCaller, const char *args);
void moveBangCommand(HWND hCaller, const char *args);

void BangBoxHook(HWND hCaller, const char *args);
void linesBangCommand(HWND hCaller, const char *args);

void shrinkBangCommand(HWND hCaller, const char *args);
void growBangCommand(HWND hCaller, const char *args);

void moveByBangCommand(HWND hCaller, const char *args);

void scrollFlashOnByBangCommand(HWND hCaller, const char *args);
void scrollFlashOffByBangCommand(HWND hCaller, const char *args);
void scrollFlashToggleByBangCommand(HWND hCaller, const char *args);

void hideTasksToggleBangCommand(HWND hCaller, const char *args);
void hideTasksOnBangCommand(HWND hCaller, const char *args);
void hideTasksOffBangCommand(HWND hCaller, const char *args);

//void testBangCommand(HWND hCaller, const char *args);

// undocumented SwitchToThisWindow API (maybe this should be in LSAPI?)
void (WINAPI *SwitchToThisWindow)(HWND, int);

const char aboutStr[] = "Taskbar3.dll: 0.301 beta 2 (jesus_mjjg)";

Taskbar *taskbar;

//Jesus
vector<string> NoFlashList;
vector<string> RemoveList;
vector<string> HideList;
vector<string> AddList;
vector<string> NoScrollList;
//Jesus

BOOL IsSubString(string big, string small)
{
	string::iterator itBig = big.begin();
	string::iterator itSmall = small.begin();

	while(itBig != big.end() && itSmall != small.end()){
		if(*itBig == *itSmall)
			++itSmall;
        ++itBig;
	}
	return(itSmall == small.end());
}

BOOL IsOnTaskbarList(HWND hWnd, vector<string> theList){
	char tmpTitle[BUFFER_SIZE], tmpClass[BUFFER_SIZE];

	GetWindowText(hWnd, tmpTitle, 1024);
	GetClassName(hWnd, tmpClass, 1024);
    vector<string>::iterator it;

	for (it = theList.begin();it != theList.end();++it){
		if (IsSubString(tmpTitle,*it) || IsSubString(tmpClass,*it))
			return TRUE;
	}
	return FALSE;
}

int initModuleEx(HWND hParent, HINSTANCE hInstance, LPCSTR pszPath)
{
	InitCommonControls();

	SwitchToThisWindow = (void (WINAPI *)(HWND, int)) GetProcAddress(
	                         GetModuleHandle("USER32.DLL"),
	                         "SwitchToThisWindow");

	taskbar = new Taskbar;
	taskbar->OnLoad(hInstance);

	AddBangCommand("!TaskbarHide", hideBangCommand);
	AddBangCommand("!TaskbarToggle", toggleBangCommand);
	AddBangCommand("!TaskbarShow", showBangCommand);

	AddBangCommand("!TaskbarSize", sizeBangCommand);
	AddBangCommand("!TaskbarMove", moveBangCommand);

	AddBangCommand("!TaskbarBoxHook", BangBoxHook);
	AddBangCommand("!TaskbarLines", linesBangCommand);
	AddBangCommand("!TaskbarShrink", shrinkBangCommand);
	AddBangCommand("!TaskbarGrow", growBangCommand);
	AddBangCommand("!TaskbarMoveBy", moveByBangCommand);

	AddBangCommand("!TaskbarScrollFlashOn", scrollFlashOnByBangCommand);
	AddBangCommand("!TaskbarScrollFlashOff", scrollFlashOffByBangCommand);
	AddBangCommand("!TaskbarScrollFlashToggle", scrollFlashToggleByBangCommand);

	AddBangCommand("!TaskbarToggleTasks", hideTasksToggleBangCommand);
	AddBangCommand("!TaskbarHideTasks", hideTasksOnBangCommand);
	AddBangCommand("!TaskbarShowTasks", hideTasksOffBangCommand);

//	AddBangCommand("!TaskbarTest", testBangCommand);

	return 0;
}

void quitModule(HINSTANCE hInstance)
{
	RemoveBangCommand("!TaskbarHide");
	RemoveBangCommand("!TaskbarToggle");
	RemoveBangCommand("!TaskbarShow");

	RemoveBangCommand("!TaskbarSize");
	RemoveBangCommand("!TaskbarMove");
	
	RemoveBangCommand("!TaskbarBoxHook");
    RemoveBangCommand("!TaskbarLines");

	RemoveBangCommand("!TaskbarShrink");
	RemoveBangCommand("!TaskbarGrow");

	RemoveBangCommand("!TaskbarMoveBy");

	RemoveBangCommand("!TaskbarScrollFlashOn");
	RemoveBangCommand("!TaskbarScrollFlashOff");
	RemoveBangCommand("!TaskbarScrollFlashToggle");

	RemoveBangCommand("!TaskbarToggleTasks");
	RemoveBangCommand("!TaskbarHideTasks");
	RemoveBangCommand("!TaskbarShowTasks");

	//RemoveBangCommand("!TaskbarTest");

	taskbar->OnUnload();
	delete taskbar;
}

extern "C" int WINAPI DllMain(HINSTANCE hInstance, DWORD dwReason, LPVOID pvReserved)
{
	if (dwReason == DLL_PROCESS_ATTACH)
		DisableThreadLibraryCalls(hInstance);

	return 1;
}


// bang commands
//

/*void testBangCommand(HWND hCaller, const char *args)
{

}*/

void hideBangCommand(HWND hCaller, const char *args)
{
	taskbar->hide();
}

void toggleBangCommand(HWND hCaller, const char *args)
{
	if (taskbar->isVisible())
		taskbar->hide();
	else
		taskbar->show();
}

void showBangCommand(HWND hCaller, const char *args)
{
	taskbar->show();
}

void Taskbar::OnMove( int xP, int yP )
{
	taskbarX = xP;
	taskbarY = yP;
    SetWindowPosition();
}

void StringToInts(const char *args, int *int1, int *int2){
    //code from www.mindjunction.com
    char token1[BUFFER_SIZE], token2[BUFFER_SIZE], extra_text[BUFFER_SIZE];
    char* tokens[11];
    int count;
    tokens[0] = token1;
    tokens[1] = token2; 

    count = LCTokenize (args, tokens, 2, extra_text);

    *int1 = atoi(token1);
    *int2 = atoi(token2);
}

void sizeBangCommand(HWND hCaller, const char *args){
    //code from www.mindjunction.com
	int cx = 0, cy = 0;
	StringToInts(args, &cx, &cy);
	if (cx <= 0)
		cx = taskbar->width;
	if (cy <= 0)
		cy = taskbar->height;

	taskbar->savedWidth = cx;
	MoveWindow( taskbar->getHwnd(),  taskbar->taskbarX, taskbar->taskbarY, cx, cy, TRUE);
}

void moveBangCommand(HWND hCaller, const char *args){
    //code from www.mindjunction.com
	int cx = 0, cy = 0;
	StringToInts(args, &cx, &cy);

	char tmpClass[BUFFER_SIZE];
    GetClassName(GetParent(taskbar->getHwnd()), tmpClass, 1024);

	if (!lstrcmpi(tmpClass,"LsBoxClass") && taskbar->LsBoxed){
		RECT r;
		GetClientRect( GetParent(taskbar->getHwnd()), &r );
		if (cx < 0)
			cx = (r.right - r.left) + cx;
		if (cy < 0)
			cy = (r.bottom - r.top) + cy;
	}
	else {
		if (cx < 0)
			cx = SCREEN_WIDTH + cx;
		if (cy < 0)
			cy = SCREEN_HEIGHT + cy;
	}

	MoveWindow( taskbar->getHwnd(),  cx, cy, taskbar->width, taskbar->height, TRUE);
}

void linesBangCommand(HWND hCaller, const char *args){
    //code from www.mindjunction.com
	int lines = 0;
    char token1[BUFFER_SIZE], extra_text[BUFFER_SIZE];
    char* tokens[11];
    int count;
    tokens[0] = token1;

    count = LCTokenize (args, tokens, 1, extra_text);

    lines = atoi(token1);

	taskbar->taskbarLines =  lines;
	if (taskbar->taskbarLines <= 0)
	  taskbar->taskbarLines = 1;

    taskbar->Layout();

}

void shrinkBangCommand(HWND hCaller, const char *args){
    //code from www.mindjunction.com
	int cx = 0, cy = 0;
	StringToInts(args, &cx, &cy);
	if (cx <= 0)
		cx = 0;
	if (cy <= 0)
		cy = 0;
	taskbar->savedWidth = taskbar->width - cx;
	MoveWindow( taskbar->getHwnd(),  taskbar->taskbarX, taskbar->taskbarY, taskbar->width - cx,taskbar->height - cy, TRUE);
}

void growBangCommand(HWND hCaller, const char *args){
    //code from www.mindjunction.com
	int cx = 0, cy = 0;
	StringToInts(args, &cx, &cy);
	if (cx <= 0)
		cx = 0;
	if (cy <= 0)
		cy = 0;
	taskbar->savedWidth = taskbar->width + cx;
	MoveWindow( taskbar->getHwnd(),  taskbar->taskbarX, taskbar->taskbarY, taskbar->width + cx,taskbar->height + cy, TRUE);
}

void moveByBangCommand(HWND hCaller, const char *args){
    //code from www.mindjunction.com
	int cx = 0, cy = 0;
	StringToInts(args, &cx, &cy);

	MoveWindow( taskbar->getHwnd(),  taskbar->taskbarX + cx, taskbar->taskbarY + cy, taskbar->width ,taskbar->height, TRUE);
}

void scrollFlashOnByBangCommand(HWND hCaller, const char *args){
	taskbar->timerScrollEnabled = true;
	taskbar->scrollWas = (GetKeyState(VK_SCROLL) != 0);
}

void scrollFlashOffByBangCommand(HWND hCaller, const char *args){
	taskbar->timerScrollEnabled = false;
}

void scrollFlashToggleByBangCommand(HWND hCaller, const char *args){
	if(taskbar->timerScrollEnabled)
		taskbar->timerScrollEnabled = false;
	else
		taskbar->timerScrollEnabled = true;
}

void hideTasksToggleBangCommand(HWND hCaller, const char *args){
	if(taskbar->reallyHidden)
		hideTasksOffBangCommand(NULL, "");		
	else
		hideTasksOnBangCommand(NULL, "");
}

void hideTasksOnBangCommand(HWND hCaller, const char *args){
	taskbar->reallyHidden = true;
}
void hideTasksOffBangCommand(HWND hCaller, const char *args){
	taskbar->reallyHidden = false;
}

// utility functions
//

HFONT GetRCFont(const char *name, const LOGFONT *lfDefault)
{
	LOGFONT lf;
	char configName[32];

	if (GetRCBool(name, TRUE))
	{
		memset(&lf, 0, sizeof(LOGFONT));

		wsprintf(configName, "%s", name);
		GetRCString(configName, lf.lfFaceName, lfDefault->lfFaceName, LF_FACESIZE);

		wsprintf(configName, "%sHeight", name);
		lf.lfHeight = GetRCInt(configName, lfDefault->lfHeight);

		wsprintf(configName, "%sBold", name);
		lf.lfWeight = GetRCBool(configName, TRUE) ? FW_BOLD : FW_NORMAL;

		wsprintf(configName, "%sItalic", name);
		lf.lfItalic = GetRCBool(configName, TRUE);

		wsprintf(configName, "%sUnderline", name);
		lf.lfUnderline = GetRCBool(configName, TRUE);
	}
	else
	{
		memcpy(&lf, lfDefault, sizeof(LOGFONT));
	}

	return CreateFontIndirect(&lf);
}

BOOL IsAppWindow( HWND hWnd )
{
	if( IsOnTaskbarList(hWnd, RemoveList) )
		return FALSE;

	if( IsOnTaskbarList(hWnd, AddList) )
		return TRUE;

	if ( !IsWindowVisible( hWnd ) )
		return FALSE;

	if ( GetWindow( hWnd, GW_OWNER ) )
		return FALSE;

	LONG styleEx = GetWindowLong( hWnd, GWL_EXSTYLE );

	if ( styleEx & WS_EX_APPWINDOW )
		return TRUE;

	if ( styleEx & WS_EX_TOOLWINDOW )
		return FALSE;

	if ( GetWindowLong( hWnd, GWL_USERDATA ) == 0x49474541 )
		return FALSE;

	return TRUE;
}

// from jugg's tasks.dll, based on code from Fahim and re5ource
HICON GetIconFromWindow( HWND hWnd, BOOL bBigIcon )
{
	HICON hIcon = NULL;

	if (hWnd)
	{
		if ( bBigIcon )
		{
			SendMessageTimeout( hWnd, WM_GETICON, ICON_BIG, 0, SMTO_ABORTIFHUNG, 1000, (LPDWORD) &hIcon );
			if ( !hIcon )
				hIcon = (HICON) GetClassLong( hWnd, GCL_HICON );
			if ( !hIcon )
				SendMessageTimeout( hWnd, WM_QUERYDRAGICON, 0, 0, SMTO_ABORTIFHUNG, 1000, (LPDWORD) &hIcon );
			if ( !hIcon )
				SendMessageTimeout( hWnd, WM_GETICON, ICON_SMALL, 0, SMTO_ABORTIFHUNG, 1000, (LPDWORD) &hIcon );
			//if( !hIcon ) hIcon = (HICON) GetClassLong( hWnd, GCL_HICONSM );
			if ( !hIcon )
				hIcon = (HICON) GetClassLong( hWnd, GCL_HICON );
		}
		else
		{
			SendMessageTimeout( hWnd, WM_GETICON, ICON_SMALL, 0, SMTO_ABORTIFHUNG, 1000, (LPDWORD) &hIcon );
			if ( !hIcon )
				hIcon = (HICON) GetClassLong( hWnd, GCL_HICONSM );
			if ( !hIcon )
				SendMessageTimeout( hWnd, WM_QUERYDRAGICON, 0, 0, SMTO_ABORTIFHUNG, 1000, (LPDWORD) &hIcon );
			if ( !hIcon )
				SendMessageTimeout( hWnd, WM_GETICON, ICON_BIG, 0, SMTO_ABORTIFHUNG, 1000, (LPDWORD) &hIcon );
			//if( !hIcon ) hIcon = (HICON) GetClassLong( hWnd, GCL_HICON );
			if ( !hIcon )
				hIcon = (HICON) GetClassLong( hWnd, GCL_HICONSM );
		}
	}

	return hIcon;
}

//
// Taskbar
//

Taskbar::Taskbar()
{
	hWnd = NULL;
	hInstance = NULL;
	hToolTips = NULL;

	edge = EDGE_BOTTOM;
	height = 0;
	width = 0;

	neededHeight = 0;

	screenHeight = 0;
	screenWidth = 0;

	active = FALSE;
	mouseOver = FALSE;
	capture = NULL;

	startButton = FALSE;
	startMousePressed = FALSE;
	startMouseOver = FALSE;

	skin = NULL;
	buttonSkin = NULL;
	activeButtonSkin = NULL;
	flashSkin = NULL;
	minimizedSkin = NULL;

	activeFont = 0;
	font = 0;
	minimizedFont = 0;
	lastActive = GetActiveWindow();
	//lastActiveCount = 0;
	taskbarLines = 1;
	flashCount = 0;
	scrollWas = (GetKeyState(VK_SCROLL) != 0);
	alwaysOnTop = FALSE;
	savedWidth = 0;
	LsBoxed = false;
	timerScrollEnabled = false;
	reallyHidden = false;

//	buttonFlashState = true;
}

Taskbar::~Taskbar()
{
	// ...
}

UINT messages[] = {
                      LM_LSSELECT,
                      LM_GETMINRECT,
                      LM_REDRAW,
                      LM_WINDOWACTIVATED,
                      LM_WINDOWCREATED,
                      LM_WINDOWDESTROYED,
					  LM_GETREVID,
                      0
                  };

void Taskbar::OnLoad( HINSTANCE hInstanceP )
{
	hInstance = hInstanceP;

	// get config variables from step.rc
	ReadConfig();

	// register taskbar window class
	WNDCLASSEX wc;
	memset( &wc, 0, sizeof(WNDCLASSEX) );

	wc.cbSize = sizeof(WNDCLASSEX);
	wc.style = CS_GLOBALCLASS;
	wc.lpfnWndProc = Taskbar::WndProc;
	wc.cbWndExtra = sizeof(Taskbar *);
	wc.hInstance = hInstance;
	wc.hCursor = LoadCursor( NULL, IDC_ARROW );
	wc.lpszClassName = WC_TASKBAR;

	if ( !RegisterClassEx( &wc ) )
	{
		RESOURCE_MSGBOX( hInstance, IDS_TASKBAR_ERROR1,
		                 "Could not register taskbar window class.", "Taskbar")
		return ;
	}

	// create the taskbar window
	hWnd = CreateWindowEx( alwaysOnTop ? WS_EX_TOPMOST |WS_EX_TOOLWINDOW : WS_EX_TOOLWINDOW,
	                       WC_TASKBAR,
	                       NULL,
						   //alwaysOnTop ? WS_POPUP : WS_CHILD,
						   WS_POPUP,
	                       0, 0, 0, 0,
	                       //GetLitestepWnd(), 20020504: disappearing taskbar problem
						   NULL,
	                       NULL,
	                       hInstance,
	                       (LPVOID) this );

    //drag and drop support
	//DragAcceptFiles(hWnd, true);

	if ( !hWnd )
	{
		RESOURCE_MSGBOX( hInstance, IDS_TASKBAR_ERROR2,
		                 "Could not create taskbar window.", "Taskbar")

		return ;
	}

	// create tooltips control
	hToolTips = CreateWindowEx( WS_EX_TOOLWINDOW | WS_EX_TOPMOST,
	                            TOOLTIPS_CLASS,
	                            NULL,
	                            TTS_ALWAYSTIP | WS_POPUP,
	                            0, 0, 0, 0,
	                            hWnd,
	                            NULL,
	                            hInstance,
	                            NULL );

	// adjust the Z-order, size, position
	SetWindowPos( hWnd, HWND_TOP, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE | SWP_NOZORDER);
    
	// hide us from the VWM
	SetWindowLong( hWnd, GWL_USERDATA, magicDWord );

	// create font and compute needed height
	OnSettingChange();//666 devrait se trouver dans ReadConfig()

	// show the window
	visible = true;
	ShowWindow(hWnd, SW_SHOWNOACTIVATE);

	// enumerate current tasks, setup buttons and start update timer
	EnumWindows(Taskbar::EnumWindowsProc, (LPARAM) this);
	Layout();

	SetTimer(hWnd, TIMER_UPDATE, timerRefresh, NULL);//will crash sometimes
//	if(timerScrollEnabled)
	//so setting it on only by bang will work
	SetTimer(hWnd, TIMER_UPDATE_SCROLL, timerScroll, NULL);//will crash sometimes
	// register for LS messages
	SendMessage( GetLitestepWnd(), LM_REGISTERMESSAGE,
	             (WPARAM) hWnd, (LPARAM) ::messages );
}

void Taskbar::OnUnload()
{
	// shutdown timer
	KillTimer(hWnd, TIMER_UPDATE);
	KillTimer(hWnd, TIMER_UPDATE_SCROLL);

	// unregister messages
	SendMessage( GetLitestepWnd(), LM_UNREGISTERMESSAGE,
	             (WPARAM) hWnd, (LPARAM) ::messages );

	// destroy tooltips control
	if ( hToolTips )
		DestroyWindow( hToolTips );

	// destroy window and unregister class
	if ( hWnd )
		DestroyWindow( hWnd );

	UnregisterClass( WC_TASKBAR, hInstance );

	// release skins
	delete skin;
	delete buttonSkin;
	delete activeButtonSkin;
	delete minimizedSkin;
	delete flashSkin;

	// release fonts
	DeleteObject(activeFont);
	DeleteObject(font);
	DeleteObject(minimizedFont);
	DeleteObject(activeFont);

	// release buttons
	for (TaskbarButtonIterator it = buttons.begin(); it != buttons.end(); ++it)
	{
		delete (*it);
		*it = NULL;
	}

	buttons.clear();
}

void Taskbar::hide()
{
	if (visible)
	{
		visible = false;
		ShowWindow(hWnd, SW_HIDE);
	}
}

void Taskbar::show()
{
	if (!visible)
	{
		visible = true;
		ShowWindow(hWnd, SW_SHOWNOACTIVATE);
	}
}

void Taskbar::Invalidate( int x, int y, int cx, int cy )
{
	RECT r;

	r.left = x;
	r.top = y;
	r.right = x + cx;
	r.bottom = y + cy;

	InvalidateRect( hWnd, &r, FALSE );
}

TaskbarButton *Taskbar::GetCapture()
{
	return capture;
}

void Taskbar::SetCapture( TaskbarButton *button )
{
	::SetCapture( hWnd );
	capture = button;
}

void Taskbar::ReleaseCapture()
{
	::ReleaseCapture();
	capture = NULL;
}

TaskbarButton *Taskbar::ButtonAtPoint( int x, int y )
{
	for (TaskbarButtonIterator it = buttons.begin(); it != buttons.end(); ++it)
	{
		if ( (*it)->HasPoint(x, y) )
		{
			return (*it);
		}
	}

	return NULL;
}

int Taskbar::GetRCMouse(const char * rcString, int intDefault){
	char buffer[MAX_PATH];

	if ( GetRCString( rcString, buffer, "", 16 ) ){
		if ( !stricmp( buffer, ".select" ) )
			return MOUSE_SELECT;
		else if ( !stricmp( buffer, ".menu" ) )
			return MOUSE_MENU;
		else if ( !stricmp( buffer, ".close" ) )
			return MOUSE_CLOSE;
		else if ( !stricmp( buffer, ".close2" ) )
			return MOUSE_CLOSE2;
		else if ( !stricmp( buffer, ".ontop" ) )
			return MOUSE_ONTOP;
		else
			return MOUSE_SELECT;
	}
	else
		return intDefault;
}

void Taskbar::GetRCList(char *text, vector<string> &theList){
	//code from www.mindjunction.com
	FILE *f;
	f = LCOpen (NULL);
	if (f){
		char buffer[BUFFER_SIZE];
 		char token1[1204], token2[BUFFER_SIZE], extra_text[BUFFER_SIZE];
 		char* tokens[2];
	 
 		tokens[0] = token1;
 		tokens[1] = token2;
 
	 	buffer[0] = 0;

		while (LCReadNextConfig (f, text, buffer, sizeof (buffer)))
		{
			int count;
 
			token1[0] = token2[0] = extra_text[0];
  
 			count = LCTokenize (buffer, tokens, 2, extra_text);
  
  			// token 1 is "*TaskbarXXXX", token 2 is the
  			// application class or title, and the extra_text are something to forget.
  			if (count >= 2)
				theList.push_back(tokens[1]);
 		}
	}
}

void Taskbar::Get1To4Settings(char *text, int &left, int &right, int &top, int &bottom, int dv_left, int dv_right, int dv_top, int dv_bottom){
	char buffer[BUFFER_SIZE];
 	char token1[1204], token2[BUFFER_SIZE], token3[BUFFER_SIZE], token4[BUFFER_SIZE], extra_text[BUFFER_SIZE];
 	char* tokens[4];
	 
 	tokens[0] = token1;
 	tokens[1] = token2;
 	tokens[2] = token3;
 	tokens[3] = token4;
 
	buffer[0] = 0;

    if(GetRCLine(text, buffer, 1024, "")){
		int count;
 
		token1[0] = token2[0] = token3[0] = token4[0] = extra_text[0];
  
	 	count = LCTokenize (buffer, tokens, 4, extra_text);
        if(count >= 1){
			if(count >=4){
				left = atoi(token1);
				right = atoi(token2);
				top = atoi(token3);
				bottom = atoi(token4);
			}
			else{
				right = top = bottom = left = atoi(token1);
			}
		}
		else{
			right = dv_right;
			top = dv_top;
			bottom = dv_bottom;
			left = dv_left;
		}
	}
	else{
		right = dv_right;
		top = dv_top;
		bottom = dv_bottom;
		left = dv_left;
	}
}

void Taskbar::ReadConfig()
{
	char buffer[MAX_PATH];
	char bufferLeft[MAX_PATH];
	char bufferRight[MAX_PATH];

	// flash
	isTaskFlash = GetRCBool( "TaskButtonFlash", TRUE);

	// edge
	if ( GetRCString( "TaskbarEdge", buffer, "", 16 ) )
	{
		edge = EDGE_BOTTOM;
		if ( !stricmp( buffer, "TOP" ) )
			edge = EDGE_TOP;
	}

	// appearance
	msTaskbar = GetRCBool( "TaskbarMSColors", TRUE );
	foreColor1 = GetRCColor( "TaskbarFore", RGB( 192, 192, 192 ) );
	foreColor2 = GetRCColor( "TaskbarFore2", RGB( 64, 64, 64 ) );
	backColor = GetRCColor( "TaskbarBack", RGB( 128, 128, 128 ) );
	activeColor = GetRCColor( "TaskbarActive", RGB( 128, 128, 128 ) );
	minColor = GetRCColor( "TaskbarMin", RGB( 128, 128, 128 ) );
	//textColor = GetRCColor( "TaskbarText", msTaskbar ? GetSysColor( COLOR_BTNTEXT ) : RGB( 255, 255, 255 ) );
	flashColor = GetRCColor( "TaskbarFlash", RGB( 224, 224, 224 ) );

	// skin settings
	noSkinShift = GetRCBool( "TaskbarNoSkinShift", TRUE );
	noTextShift = GetRCBool( "TaskbarNoTextShift", TRUE );

	// taskbar skin
	if ( GetRCString( "TaskbarSkin", buffer, "", MAX_PATH ) )
	{
		GetRCString( "TaskbarLeft", bufferLeft, "", MAX_PATH );
		GetRCString( "TaskbarRight", bufferRight, "", MAX_PATH );

		skin = new TaskbarSkin( bufferLeft, buffer, bufferRight );
	}

	// taskbar button skin
	if ( GetRCString( "TaskButtonSkin", buffer, "", MAX_PATH ) )
	{
		GetRCString( "TaskButtonLeft", bufferLeft, "", MAX_PATH );
		GetRCString( "TaskButtonRight", bufferRight, "", MAX_PATH );

		buttonSkin = new TaskbarSkin( bufferLeft, buffer, bufferRight );
	}

	// taskbar button flash skin
	if ( GetRCString( "TaskButtonFlashSkin", buffer, "", MAX_PATH ) )
	{
		GetRCString( "TaskButtonFlashLeft", bufferLeft, "", MAX_PATH );
		GetRCString( "TaskButtonFlashRight", bufferRight, "", MAX_PATH );

		flashSkin = new TaskbarSkin( bufferLeft, buffer, bufferRight );
	}

	// active taskbar button skin
	if ( GetRCString( "TaskButtonSkinActive", buffer, "", MAX_PATH ) )
	{
		GetRCString( "TaskButtonLeftActive", bufferLeft, "", MAX_PATH );
		GetRCString( "TaskButtonRightActive", bufferRight, "", MAX_PATH );

		activeButtonSkin = new TaskbarSkin( bufferLeft, buffer, bufferRight );
	}

	NONCLIENTMETRICS ncm, ncm2, ncm3, ncm4;
	ncm4.cbSize = ncm3.cbSize = ncm2.cbSize = ncm.cbSize = sizeof(NONCLIENTMETRICS);

	SystemParametersInfo(SPI_GETNONCLIENTMETRICS, sizeof(NONCLIENTMETRICS), &ncm, 0);
	SystemParametersInfo(SPI_GETNONCLIENTMETRICS, sizeof(NONCLIENTMETRICS), &ncm2, 0);
	SystemParametersInfo(SPI_GETNONCLIENTMETRICS, sizeof(NONCLIENTMETRICS), &ncm3, 0);
	SystemParametersInfo(SPI_GETNONCLIENTMETRICS, sizeof(NONCLIENTMETRICS), &ncm4, 0);
	
	font = GetRCFont("TaskbarFont", &ncm.lfCaptionFont);
	activeFont = GetRCFont("TaskbarActiveFont", &ncm2.lfCaptionFont);
	minimizedFont = GetRCFont("TaskbarMinFont", &ncm3.lfCaptionFont);
	flashFont = GetRCFont("TaskbarFlashFont", &ncm4.lfCaptionFont);

	textColor = GetRCColor( "TaskbarText", msTaskbar ? GetSysColor( COLOR_BTNTEXT ) : RGB( 255, 255, 255 ) );
	textActiveColor = GetRCColor( "TaskbarTextActive", msTaskbar ? GetSysColor( COLOR_BTNTEXT ) : RGB( 0, 0, 0 ) );
    textMinimizedColor = GetRCColor( "TaskbarTextMin", msTaskbar ? GetSysColor( COLOR_BTNTEXT ) : RGB( 0, 0, 0 ) );
    textFlashColor = GetRCColor( "TaskbarTextFlash", msTaskbar ? GetSysColor( COLOR_BTNTEXT ) : RGB( 255, 255, 255 ) );

	taskbarPositioning = GetRCBoolDef("TaskbarPositioning", FALSE);
	if (taskbarPositioning)
	{
		taskbarX = GetRCInt("TaskbarX", 0);
		taskbarXS = taskbarX;
		if (taskbarX < 0)
			taskbarX = SCREEN_WIDTH + taskbarX;
		taskbarY = GetRCInt("TaskbarY", 0);
		taskbarYS = taskbarY;
		if (taskbarY < 0)
			taskbarY = SCREEN_HEIGHT + taskbarY;
		savedX = taskbarX;
	}

	// SEE: OnSettingChange for width setting TaskbarWidth

	alwaysOnTop = GetRCBool("TaskbarAlwaysOnTop", TRUE);
	reallyHidden = GetRCBool("TaskbarHideTasks", TRUE);
	
	taskbarHeight = GetRCInt( TEXT("TaskBarHeight"), 0 );

	screenWidth = GetRCInt( TEXT("TaskBarWidth"), SCREEN_WIDTH );
	savedWidth = screenWidth;
	
	maxTaskWidth = GetRCInt( TEXT("TaskbarMaxTaskWidth"), SCREEN_WIDTH );

	Get1To4Settings("TaskbarBorder", borderLeft, borderRight, borderTop, borderBottom, 4, 4, 4, 4);
    /**/	
    /*borderLeft = GetRCInt( TEXT("TaskBarBorderLeft"), 4 );
	borderTop = GetRCInt( TEXT("TaskBarBorderTop"), 4 );
	borderRight = GetRCInt( TEXT("TaskBarBorderRight"), 4 );
	borderBottom = GetRCInt( TEXT("TaskBarBorderBottom"), 4 );*/
    /**/
    /**/
	Get1To4Settings("TaskbarButtonBorder", buttonBorderLeft, buttonBorderRight, buttonBorderTop, buttonBorderBottom, 4, 4, 4, 4);
	/*buttonBorderLeft = GetRCInt( TEXT("TaskBarButtonBorderLeft"), 24 );
	buttonBorderTop = GetRCInt( TEXT("TaskBarButtonBorderTop"), 4 );
	buttonBorderRight = GetRCInt( TEXT("TaskBarButtonBorderRight"), 4 );
	buttonBorderBottom = GetRCInt( TEXT("TaskBarButtonBorderBottom"), 4 );*/
    /**/

	Get1To4Settings("TaskbarButtonTextBorder", buttonTextBorderLeft, buttonTextBorderRight, buttonTextBorderTop, buttonTextBorderBottom, 20, 0, 0, 0);

	spacing = GetRCInt( TEXT("TaskBarSpacing"), 3 );
	//iconSpacing = GetRCInt( TEXT("TaskBarIconSpacing"), 3 );
	iconSize = GetRCInt( TEXT("TaskBarIconSize"), 16);

	// minimized taskbar button skin
	if ( GetRCString( "TaskButtonSkinMin", buffer, "", MAX_PATH ) )
	{
		GetRCString( "TaskButtonLeftMin", bufferLeft, "", MAX_PATH );
		GetRCString( "TaskButtonRightMin", bufferRight, "", MAX_PATH );

		minimizedSkin = new TaskbarSkin( bufferLeft, buffer, bufferRight );
	}

	showIcon = GetRCBoolDef("TaskbarIconShow", TRUE) ? true : false;

    /*mouse buttons on bar*/
	GetRCLine("TaskBarLButtonDown", taskbarLButtonDown, MAX_LINE_LENGTH, "!none");
	GetRCLine("TaskBarRButtonDown", taskbarRButtonDown, MAX_LINE_LENGTH, "!none");
	GetRCLine("TaskBarMButtonDown", taskbarMButtonDown, MAX_LINE_LENGTH, "!none");
	GetRCLine("TaskBarLButtonUp", taskbarLButtonUp, MAX_LINE_LENGTH, "!none");
	GetRCLine("TaskBarRButtonUp", taskbarRButtonUp, MAX_LINE_LENGTH, "!none");
	GetRCLine("TaskBarMButtonUp", taskbarMButtonUp, MAX_LINE_LENGTH, "!none");

	/*action on tiles*/
	TaskBarTileLButton = GetRCMouse("TaskBarTButtonL", MOUSE_SELECT);
	TaskBarTileMButton = GetRCMouse("TaskBarTButtonM", MOUSE_NONE);
	TaskBarTileRButton = GetRCMouse("TaskBarTButtonR", MOUSE_MENU);

	taskbarLines =  GetRCInt( TEXT("TaskBarLines"), 1 );
	if (taskbarLines <= 0)
	  taskbarLines = 1;

	timerRefresh =  GetRCInt( TEXT("TaskBarRefresh"), 500 );
    if (timerRefresh < 100)
		timerRefresh = 500;
	timerScroll =  GetRCInt( TEXT("TaskBarScrollFlash"), 500 );
    if (timerScroll < 100)
		timerScroll = 500;
	timerScrollEnabled = GetRCBool( "TaskbarFlashScrollLock", TRUE );

	//Taskbar Lists
    /*vector<string> HideList;*/
	GetRCList("*TaskbarNoFlash", NoFlashList);
	GetRCList("*TaskbarRemove", RemoveList);
	GetRCList("*TaskbarAdd", AddList);
	GetRCList("*TaskbarNoScroll",NoScrollList);
	GetRCList("*TaskbarHide", HideList);

	iconX = GetRCInt( TEXT("TaskBarIconX"), 0);
	iconY = GetRCInt( TEXT("TaskBarIconY"), 0);

	if ( GetRCString( "TaskbarHorizontalTextAlign", buffer, "", 16 ) )
	{
		if ( !stricmp( buffer, ".center" ) )
			textX = -2;
		else if ( !stricmp( buffer, ".right" ) )
			textX = -3;
		else
			textX = -1;
	}
	else
		textX = -1;

	if ( GetRCString( "TaskbarVerticalTextAlign", buffer, "", 16 ) )
	{
		if ( !stricmp( buffer, ".center" ) )
			textY = -2;
		else if ( !stricmp( buffer, ".bottom" ) )
			textY = -3;
		else
			textY = -1;
	}
	else
		textY = -2;

	/*buttonIconBorderLeft = GetRCInt( TEXT("TaskBarButtonIconBorderLeft"), 4 );
	buttonIconBorderTop = GetRCInt( TEXT("TaskBarButtonIconBorderTop"), 4 );
	buttonIconBorderRight = GetRCInt( TEXT("TaskBarButtonIconBorderRight"), 4 );
	buttonIconBorderBottom = GetRCInt( TEXT("TaskBarButtonIconBorderBottom"), 4 );*/

	showText = GetRCBoolDef("TaskbarTextShow", TRUE) ? true : false;

	if ( GetRCString( "TaskbarButtonDirection", buffer, "", 16 ) )
	{
		if ( !stricmp( buffer, ".left" ) )
			direction = DIRECTION_LEFT;
		else
			direction = DIRECTION_RIGHT;
	}
	else
		direction = DIRECTION_RIGHT;
	//LSLogPrintf(0,"Taskbar.dll","direction %d", taskbar->direction);
}

void Taskbar::Layout()
{
	flashCount = 0;
	inCriticalProcess = true;//666

	int buttonCount = buttons.size();
	if( taskbar->reallyHidden ){
		for (TaskbarButtonIterator it = buttons.begin(); it != buttons.end(); ++it)
			if ( IsOnTaskbarList((*it)->GetTask(), HideList) )
				--buttonCount;
	}

	if (buttonCount <= 0)
	{
		InvalidateRect( hWnd, NULL, FALSE );
		inCriticalProcess = false;
		return ;
	}

	int availableWidth = width - borderLeft - borderRight;
	availableWidth = max( availableWidth, 0 );  // cannot be less than zero

	if ( buttonCount > 1 )
	{
		availableWidth = availableWidth - (buttonCount - 1) * spacing;
		availableWidth = max( availableWidth, 0 );
	}

	int widthPerButton = min(availableWidth / buttonCount,maxTaskWidth);

	int buttonHeight = (height - borderTop - borderBottom - (taskbarLines - 1) * spacing) / taskbarLines;

	int lines =0;
	int linebuttonCount;

    if(taskbarLines>1){
		if (!(buttonCount%taskbarLines))
			linebuttonCount = ((buttonCount) / taskbarLines);
		else
			linebuttonCount = ((buttonCount) / taskbarLines) + 1;
	}
	else
	  //linebuttonCount = ((buttonCount) / taskbarLines);
	  linebuttonCount = buttonCount;

    int error = availableWidth % linebuttonCount;
    int errorDone = 0;
    widthPerButton = min(availableWidth / linebuttonCount,maxTaskWidth);

	//vers la droite
	int x;
	if(taskbar->direction == DIRECTION_RIGHT)
		x = borderLeft;
	else
		x = taskbar->width - borderRight - widthPerButton;

	int y = borderTop;

	// set placement for each button
	for (TaskbarButtonIterator it = buttons.begin(); it != buttons.end(); ++it)
	{
		TaskbarButton *button = *it;
		if((taskbar->reallyHidden && !IsOnTaskbarList((*it)->GetTask(), HideList)) || !taskbar->reallyHidden){
			if ( error > 0 )
			{
				++lines;
				if((lines % taskbarLines) == 0){
			  		error--;
			  		errorDone++;
				}
 		    	button->Move( x , y, widthPerButton + 1, buttonHeight );

				if(taskbar->direction == DIRECTION_RIGHT)
					x = borderLeft + (lines / taskbarLines) * widthPerButton + (lines / taskbarLines) * spacing + errorDone;
				else 
					x = taskbar->width - borderRight - (lines / taskbarLines + 1 ) * widthPerButton - (lines / taskbarLines) * spacing - errorDone;
            	y = borderTop + (lines % taskbarLines) * buttonHeight + (lines % taskbarLines) * spacing;
			}
			else
			{
				button->Move( x , y, widthPerButton, buttonHeight );
				++lines;

				if(taskbar->direction == DIRECTION_RIGHT)
					x = borderLeft + (lines / taskbarLines) * widthPerButton + (lines / taskbarLines) * spacing + errorDone;
				else
					x = taskbar->width - borderRight - (lines / taskbarLines + 1) * widthPerButton - (lines / taskbarLines) * spacing - errorDone;
				y = borderTop + (lines % taskbarLines) * buttonHeight + (lines % taskbarLines) * spacing;
			}
		}
		else
			button->Move( x , y, 0, 0);
	}

	InvalidateRect( hWnd, NULL, FALSE );
	inCriticalProcess = false;//666
}



void Taskbar::SetWindowPosition()
{
	if (taskbarPositioning)
	{
		SetWindowPos( hWnd,  NULL, taskbarX, taskbarY,screenWidth ,neededHeight, SWP_NOACTIVATE | SWP_NOZORDER );
	}
	else
	{
		if ( edge == EDGE_TOP )
		{
			SetWindowPos( hWnd, NULL, 0, 0, screenWidth, neededHeight, SWP_NOACTIVATE | SWP_NOZORDER );
		}
		else
		{
			SetWindowPos( hWnd, NULL, 0, screenHeight - neededHeight, screenWidth, neededHeight, SWP_NOACTIVATE | SWP_NOZORDER );
		}
	}
}

TaskbarButton *Taskbar::SearchForButtonByHandle( HWND hWnd )
{
	for (TaskbarButtonIterator it = buttons.begin(); it != buttons.end(); ++it)
	{
		if ( hWnd == (*it)->GetTask() )
			return (*it);
	}

	return NULL;
}

void Taskbar::OnActivate( BOOL becomingActive )
{
	//what's the purpose of active?
	active = becomingActive;
}

void Taskbar::OnDisplayChange()
{
	//MoveWindow( taskbar->getHwnd(),  taskbar->taskbarX, taskbar->taskbarY, taskbar->width,taskbar->height, TRUE);
}

void Taskbar::OnGetMinRect( HWND hWnd, RECT *rectangle )//CPU 20020501
{
	TaskbarButton *button;
	button = SearchForButtonByHandle(hWnd);
	if(button)
		(button)->needRedraw = true;
}

void Taskbar::OnLButtonDown( int x, int y )
{
	TaskbarButton *button;

	if ( capture )
		button = capture;
	else
		button = ButtonAtPoint( x, y );

	if ( button )
	{
		button->OnLButtonDown( x, y );
	}
	else{
		LSExecute(NULL, Taskbar::taskbarLButtonDown, NULL);
	}
}

void Taskbar::OnMButtonDown( int x, int y )
{
	TaskbarButton *button;

	if ( capture )
		button = capture;
	else
		button = ButtonAtPoint( x, y );

	if ( button )
	{
		button->OnMButtonDown( x, y );
	}
	else{
		LSExecute(NULL, Taskbar::taskbarMButtonDown, NULL);
	}
}

void Taskbar::OnMButtonUp( int x, int y )
{
	TaskbarButton *button;

	if ( capture )
		button = capture;
	else
		button = ButtonAtPoint( x, y );

	if ( button )
		button->OnMButtonUp( x, y );
	else
	{
		LSExecute(NULL, Taskbar::taskbarMButtonUp, NULL);
	}

}


void Taskbar::OnLButtonUp( int x, int y )
{
	TaskbarButton *button;

	if ( capture )
		button = capture;
	else
		button = ButtonAtPoint( x, y );

	if ( button )
		button->OnLButtonUp( x, y );
	else{
		if( taskbar->lastActive ){
			SearchForButtonByHandle(taskbar->lastActive)->Deactivate();
			taskbar->lastActive = NULL;
		}
		LSExecute(NULL, Taskbar::taskbarLButtonUp, NULL);
	}
}

void Taskbar::OnMouseMove( int x, int y )
{
	mouseOver = TRUE;

	TaskbarButton *button;

	if ( capture )
		button = capture;
	else
		button = ButtonAtPoint( x, y );

	if ( button )
	{
		button->OnMouseMove( x, y );
		return ;
	}

	POINT pt;
	RECT r;

	pt.x = x;
	pt.y = y;

	r.left = 2;
	r.top = 4;
	r.right = r.left + 64;
	r.bottom = height - 4;

	if ( PtInRect( &r, pt ) )
		startMouseOver = TRUE;
}

void Taskbar::OnPaint( HDC hDC )
{
	HDC hdcBuffer;
	HBITMAP hbmBuffer;
	HBITMAP hbmTemp;

	hdcBuffer = CreateCompatibleDC( hDC );
	hbmBuffer = CreateCompatibleBitmap( hDC, width, height );
	hbmTemp = (HBITMAP) SelectObject( hdcBuffer, hbmBuffer );

	if ( !skin )
	{
		RECT r;
		GetClientRect( hWnd, &r );

		if ( !msTaskbar )
		{
			// simple 3D frame
			HBRUSH hbrBack;
			HPEN hpnFore;
			HPEN hpnTemp;

			hbrBack = CreateSolidBrush( backColor );
			FillRect( hdcBuffer, &r, hbrBack );
			DeleteObject( hbrBack );

			hpnFore = CreatePen( PS_SOLID, 1, foreColor2 );
			hpnTemp = (HPEN) SelectObject( hdcBuffer, hpnFore );

			MoveToEx( hdcBuffer, r.left, r.bottom - 1, NULL );
			LineTo( hdcBuffer, r.right - 1, r.bottom - 1 );
			LineTo( hdcBuffer, r.right - 1, r.top );

			SelectObject( hdcBuffer, hpnTemp );
			DeleteObject( hpnFore );

			hpnFore = CreatePen( PS_SOLID, 1, foreColor1 );
			hpnTemp = (HPEN) SelectObject( hdcBuffer, hpnFore );

			LineTo( hdcBuffer, r.left, r.top );
			LineTo( hdcBuffer, r.left, r.bottom - 1 );

			SelectObject( hdcBuffer, hpnTemp );
			DeleteObject( hpnFore );
		}
		else
		{
			// MS taskbar style
			DrawEdge( hdcBuffer, &r, EDGE_RAISED, BF_RECT | BF_MIDDLE );
		}
	}
	else
	{
		// skin
		skin->Apply( hdcBuffer, 0, 0, width, height );
	}

	for (TaskbarButtonIterator it = buttons.begin(); it != buttons.end(); ++it){
	  if ((taskbar->reallyHidden ) && (!IsOnTaskbarList((*it)->GetTask(), HideList)))
		  (*it)->OnPaint( hdcBuffer );
	  else if (!taskbar->reallyHidden )
		  (*it)->OnPaint( hdcBuffer );
	}

	BitBlt( hDC, 0, 0, width, height, hdcBuffer, 0, 0, SRCCOPY );

	SelectObject( hdcBuffer, hbmTemp );
	DeleteObject( hbmBuffer );
	DeleteDC( hdcBuffer );
}

// OnRedraw
//

void Taskbar::OnRedraw( HWND hWnd, BOOL fFlash )
{
	TaskbarButton *button = SearchForButtonByHandle(hWnd);

	if (button)
	{
		button->SetFlashing(fFlash);
		button->OnUpdate(TRUE);
	}
}

void Taskbar::OnRButtonDown( int x, int y )
{
	TaskbarButton *button;

	if ( capture )
		button = capture;
	else
		button = ButtonAtPoint( x, y );

	if ( button )
		button->OnRButtonDown( x, y );
	else
      LSExecute(NULL, this->taskbarRButtonDown, NULL);

}

void Taskbar::OnRButtonUp( int x, int y )
{
	TaskbarButton *button;

	if ( capture )
		button = capture;
	else
		button = ButtonAtPoint( x, y );

	if ( button )
		button->OnRButtonUp( x, y );
	else
      LSExecute(NULL, this->taskbarRButtonUp, NULL);
}

void Taskbar::OnSettingChange()
{
	// compute needed height and set initial position
	TEXTMETRIC tm;
	HDC hDC;
	HFONT hfnTemp;

	hDC = GetDC( NULL );
	hfnTemp = (HFONT) SelectObject( hDC, font );

	GetTextMetrics( hDC, &tm );

	SelectObject( hDC, hfnTemp );
	ReleaseDC( NULL, hDC );

    
	if(taskbarHeight <= 0){
	  neededHeight = max( (const int)(tm.tmHeight + tm.tmExternalLeading), iconSize );
  	  neededHeight = neededHeight + buttonBorderTop + buttonBorderBottom;
	  neededHeight = neededHeight + borderTop + borderBottom;
	}
	else
      neededHeight = taskbarHeight;
    


	screenHeight = SCREEN_HEIGHT;

/*	if(savedWidth <= 0){
		screenWidth = GetRCInt( TEXT("TaskBarWidth"), SCREEN_WIDTH );
	    savedWidth = screenWidth;
	}
	else
		screenWidth = savedWidth;*/

	SetWindowPosition();
}

void Taskbar::OnSize( int widthP, int heightP )
{
	height = heightP;
	width = widthP;

	Layout();
}


void Taskbar::OnTimerUpdate(int whichTimer)
{

	if((whichTimer == TIMER_UPDATE_SCROLL)&&(taskbar->timerScrollEnabled)){ //6661
		if(taskbar->flashCount > 0){
			keybd_event( VK_SCROLL, 0x45, KEYEVENTF_EXTENDEDKEY | 0, 0 );
			keybd_event( VK_SCROLL, 0x45, KEYEVENTF_EXTENDEDKEY | KEYEVENTF_KEYUP, 0);
		}
		else{
			if ((scrollWas && !GetKeyState(VK_SCROLL)) || (!scrollWas && GetKeyState(VK_SCROLL))){
		  		keybd_event( VK_SCROLL, 0x45, KEYEVENTF_EXTENDEDKEY | 0, 0 );
				keybd_event( VK_SCROLL, 0x45, KEYEVENTF_EXTENDEDKEY | KEYEVENTF_KEYUP, 0);
			}
		}
		//taskbar->flashCount = 0;
	}
	else{
		if(!inCriticalProcess)//CPU 20020501
			for (TaskbarButtonIterator it = buttons.begin(); it != buttons.end(); ++it)
			{		
				(*it)->OnUpdate((*it)->needRedraw);
				(*it)->needRedraw = false;
			}

			//Layout();
	}
	//taskbar->buttonFlashState = !taskbar->buttonFlashState;
}

void Taskbar::OnWindowActivated( HWND hApplication, BOOL fFullScreen )
{
	// update buttons
	/*for (TaskbarButtonIterator it = buttons.begin(); it != buttons.end(); ++it)
		(*it)->OnUpdate();*/
	if(TaskbarButton *button = SearchForButtonByHandle(hApplication))
		(button)->needRedraw = true;//CPU 20020501
}

HWND TaskbarButton::buttonHandler(){
	return hTask;
}

BOOL Taskbar::IsAlreadyInList(HWND hWnd){
	for (TaskbarButtonIterator it = buttons.begin(); it != buttons.end(); ++it)
	{
	  if (hWnd == (*it)->buttonHandler())
		  return TRUE;
	}
    return FALSE;
}

void Taskbar::OnWindowCreated( HWND hTask )
{
	if (!IsAlreadyInList(hTask) && !IsOnTaskbarList(hTask, RemoveList)){
		TaskbarButton *button = new TaskbarButton( *this, hTask );
		button->SetSkin(buttonSkin);
		button->SetActiveSkin(activeButtonSkin);
		button->SetFlashSkin(flashSkin);
		button->SetMinimizedSkin(minimizedSkin);

		buttons.push_back( button );

		Layout();
	}
}

void Taskbar::OnWindowDestroyed( HWND hTask )
{
	for (TaskbarButtonIterator i = buttons.begin(); i != buttons.end(); ++i )
	{
		TaskbarButton *button = *i;

		if ( button->GetTask() == hTask )
		{
			delete button;
			buttons.erase( i );

			Layout();
			return ;
		}
	}
}

BOOL Taskbar::OnWindowMessage( UINT nMessage, WPARAM wParam, LPARAM lParam, LRESULT &lResult )
{
	lResult = 0;

	// relay mouse messages to tooltips control
	if ( nMessage >= WM_MOUSEFIRST && nMessage <= WM_MOUSELAST )
	{
		MSG msg;

		msg.hwnd = hWnd;
		msg.message = nMessage;
		msg.wParam = wParam;
		msg.lParam = lParam;

		SendMessage( hToolTips, TTM_RELAYEVENT, 0, (LPARAM) &msg );
	}

	switch ( nMessage )
	{
		case LM_LSSELECT:
		{
			//OnWindowActivated( NULL, FALSE );
			OnWindowActivated( (HWND) wParam, FALSE );
			return TRUE;
		}

		case LM_GETMINRECT:
		{
			OnGetMinRect( (HWND) wParam, (RECT *) lParam );
			return TRUE;
		}

		case LM_REDRAW:
		{
			OnRedraw( (HWND) wParam, (BOOL) lParam );
			return TRUE;
		}

		case LM_WINDOWACTIVATED:
		{
			OnWindowActivated( (HWND) wParam, (BOOL) lParam );
			return TRUE;
		}

		case LM_WINDOWCREATED:
		{
			OnWindowCreated( (HWND) wParam );
			/*pika*/
			/*++OnNumCount;

			if ( OnNumCount <= OnNumList.size() && OnNumExe ) {
				LSExecute(NULL, OnNumList[(OnNumCount - 1)], NULL);
			}*/
			/*pika*/

			return TRUE;
		}

		case LM_WINDOWDESTROYED:
		{
			OnWindowDestroyed( (HWND) wParam );
			/*pika*/
			/*--OnNumCount;

			if ( OnNumCount <= OnNumList.size() && OnNumExe ) {
				LSExecute(NULL, OnNumList[(OnNumCount - 1)], NULL);
			}*/
			/*pika*/
			return TRUE;
		}

		case LM_GETREVID:
		{
			char* buf = (char*)lParam;
			strcpy(buf, aboutStr);
			return strlen(buf);
		}

		case WM_ACTIVATE:
		{
			OnActivate( (LOWORD(wParam) == WA_INACTIVE) ? FALSE : TRUE );
			return TRUE;
		}

		case WM_DISPLAYCHANGE:
		{
			OnDisplayChange();
			return TRUE;
		}

		case WM_LBUTTONDOWN:
		{
			OnLButtonDown( LOWORD(lParam), HIWORD(lParam));
			return TRUE;
		}
		case WM_LBUTTONUP:
		{
			OnLButtonUp( LOWORD(lParam), HIWORD(lParam));
			return TRUE;
		}
		/*case WM_LBUTTONDBLCLK:
		{
			return TRUE;
		}*/
		case WM_MBUTTONDOWN:
		{
			OnMButtonDown( LOWORD(lParam), HIWORD(lParam));
			return TRUE;
		}
		case WM_MOUSEMOVE:
		{
			OnMouseMove( LOWORD( lParam ), HIWORD( lParam ) );
			return TRUE;
		}

		case WM_PAINT:
		{
			PAINTSTRUCT ps;
			HDC hDC;

			if ( !wParam )
				hDC = BeginPaint( hWnd, &ps );
			else
				hDC = (HDC) wParam;

			OnPaint( hDC );

			if ( !wParam )
				EndPaint( hWnd, &ps );

			return TRUE;
		}

		case WM_RBUTTONDOWN:
		{
			OnRButtonDown( (int) (short) LOWORD( lParam ), (int) (short) HIWORD( lParam ) );
			return TRUE;
		}

		case WM_RBUTTONUP:
		{
			OnRButtonUp( (int) (short) LOWORD( lParam ), (int) (short) HIWORD( lParam ) );
			return TRUE;
		}

		case WM_SETTINGCHANGE:
		{
			OnSettingChange();//666 -> jdesk -> lslog
			//LSLogPrintf(0,"Taskbar.dll","WM_SETTINGCHANGE %d", taskbar->savedWidth);
			/*taskbar->width = taskbar->savedWidth;
			taskbar->Layout();*/
			//-> ok
			return TRUE;
		}

		case WM_SIZE:
		{
			OnSize( (int) (short) LOWORD( lParam ), (int) (short) HIWORD( lParam ) );
			return TRUE;
		}

		case WM_TIMER:
		{
			switch ( wParam )
			{
				/*case TIMER_AUTOHIDE:
				{
					OnTimerAutoHide();
					return TRUE;
				}*/
				
				case TIMER_UPDATE:
				{
					OnTimerUpdate(TIMER_UPDATE);
					return TRUE;
				}
				case TIMER_UPDATE_SCROLL:
				{
					OnTimerUpdate(TIMER_UPDATE_SCROLL);
					return TRUE;
				}
			}

			return FALSE;
		}
		/*case WM_DROPFILES://faut trouver la tche et l'ouvrir avec, multiple files
		{
			LPSTR commandLine, iter;
			UINT fileCount = DragQueryFile(HDROP(wParam), ~0, NULL, 0);
			//size_t totalLength = strlen(fCommand) + 3;

			size_t totalLength = 3;
			UINT i;

			if (!fileCount)
				return FALSE;

			for (i = 0; i < fileCount; i++)
			{
				totalLength += DragQueryFile(HDROP(wParam), i, NULL, 0) + 3;
			}
			
			commandLine = new char[totalLength];
			//sprintf(commandLine, "[%s]", fCommand);
			iter = commandLine + strlen(commandLine);

			for (i = 0; i < fileCount; i++)
			{
				strcpy(iter, " \"");
				iter += strlen(iter);
				iter += DragQueryFile(HDROP(wParam), i, iter,
				totalLength - (iter - commandLine));

				*iter = '\"';
				iter++;
				*iter = '\0';
			}
			
			LSExecute(hWnd, commandLine, SW_SHOWDEFAULT);
			delete[] commandLine;
		}*/

	}

	return FALSE;
}

BOOL WINAPI Taskbar::EnumWindowsProc(HWND hWnd, LPARAM lParam)
{
	Taskbar *taskbar = (Taskbar *) lParam;

	if (taskbar && IsAppWindow(hWnd))
	{
		TaskbarButton *button = taskbar->SearchForButtonByHandle(hWnd);

		if (button == 0)
		{
			button = new TaskbarButton(*taskbar, hWnd);

			button->SetActiveSkin(taskbar->activeButtonSkin);
			button->SetSkin(taskbar->buttonSkin);
			button->SetFlashSkin(taskbar->flashSkin);
			button->SetMinimizedSkin(taskbar->minimizedSkin);

			taskbar->buttons.push_back(button);
			taskbar->listChanged = true;
		}
	}

	return TRUE;
}

LRESULT WINAPI Taskbar::WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	Taskbar *taskbar = (Taskbar *) GetWindowLong(hWnd, 0);

	if (message == WM_NCCREATE)
	{
		LPCREATESTRUCT pcs = (LPCREATESTRUCT) lParam;
		taskbar = (Taskbar *) pcs->lpCreateParams;

		taskbar->hWnd = hWnd;
		SetWindowLong(hWnd, 0, (LONG) taskbar);
	}

	if (taskbar)
	{
		LRESULT lResult;

		if (taskbar->OnWindowMessage(message, wParam, lParam, lResult))
			return lResult;
	}

	return DefWindowProc(hWnd, message, wParam, lParam);
}

//
// TaskbarButton
//

TaskbarButton::TaskbarButton( Taskbar &tb, HWND hWnd )
		: container( tb )
{
	hTask = hWnd;
	skin = NULL;
	activeskin = NULL;
	flashskin = NULL;
	minimizedskin = NULL;

	otherCaption = NULL;
	caption = NULL;

	isOnTop = FALSE;//how the fuck do i know?

//	active = (GetForegroundWindow() == hTask);
//	active = (GetFocus() == hTask);//6661
	/*if( GetFocus() != NULL)
      active = (GetFocus() == this->hTask) ;
	else*/
      active = (GetForegroundWindow() == this->hTask);

	if (hTask)
	{
		caption = GetWindowCaption();
		// add tooltip
		TOOLINFO ti;
		memset( &ti, 0, sizeof(TOOLINFO) );

		ti.cbSize = sizeof(TOOLINFO);
		ti.hwnd = container.hWnd;
		ti.uId = (UINT) this;
		ti.lpszText = caption;

		SendMessage( container.hToolTips, TTM_ADDTOOL, 0, (LPARAM) &ti );
	}
	if(taskbar->showIcon)
		icon = GetIcon();

	mousePressed = FALSE;
	mouseOver = FALSE;
	taskflash = FALSE;

	x = 0;
	y = 0;
	height = 0;
	width = 0;

	/**/
	flashing = false;
	/**/
	needRedraw = false;
}

TaskbarButton::~TaskbarButton()
{
	delete caption;
	delete otherCaption;

	// delete tooltip
	TOOLINFO ti;
	memset( &ti, 0, sizeof(TOOLINFO) );

	ti.cbSize = sizeof(TOOLINFO);
	ti.hwnd = container.hWnd;
	ti.uId = (UINT) this;

	SendMessage( container.hToolTips, TTM_DELTOOL, 0, (LPARAM) &ti );
}

BOOL TaskbarButton::HasPoint( int xP, int yP )
{
	if ( xP < (x + width) && xP >= x )
	{
		if ( yP < (y + height) && yP >= y )
			return TRUE;
	}

	return FALSE;
}

void TaskbarButton::Move( int xP, int yP, int widthP, int heightP )
{
	x = xP;
	y = yP;
	height = heightP;
	width = widthP;

	// update tooltip
	TOOLINFO ti;
	memset( &ti, 0, sizeof(TOOLINFO) );

	ti.cbSize = sizeof(TOOLINFO);
	ti.hwnd = container.hWnd;
	ti.uId = (UINT) this;
	ti.rect.left = x;
	ti.rect.top = y;
	ti.rect.right = x + width;
	ti.rect.bottom = y + height;

	SendMessage( container.hToolTips, TTM_NEWTOOLRECT, 0, (LPARAM) &ti );
}



void TaskbarButton::OnLButtonDown( int, int )
{
	switch (taskbar->TaskBarTileLButton){
		case MOUSE_SELECT:
			SelectDown();
		    break;
		case MOUSE_MENU:
			MenuDown();
		    break;
		case MOUSE_ONTOP:
			OnTopToggle();
			break;
		case MOUSE_CLOSE:
			Close();
		    break;
		case MOUSE_CLOSE2:
			Close2();
		    break;
		case MOUSE_NONE:
		    break;
        default:
		    break;
	}
}

void TaskbarButton::OnTopToggle()
{
	if (!isOnTop){
		EndDeferWindowPos(DeferWindowPos(BeginDeferWindowPos(1) ,this->hTask, HWND_TOPMOST, 0, 0, 0, 0, SWP_NOMOVE|SWP_NOSIZE));//Set ontop
		isOnTop = TRUE;
		// undocumented call to switch window activation
		SwitchToThisWindow( hTask, 1 );
		active = TRUE;
	}
	else{
		EndDeferWindowPos(DeferWindowPos(BeginDeferWindowPos(1) ,this->hTask, HWND_NOTOPMOST, 0, 0, 0, 0, SWP_NOMOVE|SWP_NOSIZE));//Set not ontop
		isOnTop = FALSE;
	}
}


void TaskbarButton::Close()
{
	SendMessage(this->hTask, WM_CLOSE,0,0);
}

void TaskbarButton::Close2()
{
	SendMessage(this->hTask, WM_DESTROY,0,0); //if cancel -> still destroyed
}

void TaskbarButton::SelectUp(){//problem when loading in a box: won't minimize
	mousePressed = FALSE;

	if ( container.GetCapture() != this ){
		return ;
	}

	container.ReleaseCapture();

	if ( !mouseOver )
	{
		container.Invalidate( x, y, width + 1, height + 1 );
		return ;
	}
    if (flashing){
	  taskbar->flashCount = taskbar->flashCount - 1;
	  //taskbar->flashCount = 0;      
	}

	/*if ( active )
	{
		// already active, minimize it
		ShowWindowAsync( hTask, SW_MINIMIZE );
		active = FALSE;
	}
	else if (taskbar->LsBoxed && taskbar->lastActive == this->hTask && !IsIconic( this->hTask )){
		char tmpClass[BUFFER_SIZE];
		GetClassName(GetParent(taskbar->getHwnd()), tmpClass, 1024);
		if (strstr(tmpClass, "LsBoxClass")){
			// already active, minimize it
			ShowWindowAsync( hTask, SW_MINIMIZE );
			active = FALSE;
		}
	}
	else
	{
		if ( !IsIconic( hTask ))
		{
			// if not minimized, move to application's current desktop
			SendMessage( GetLitestepWnd(), LM_BRINGTOFRONT, 0, (LPARAM) hTask );
		}

		// undocumented call to switch window activation
		SwitchToThisWindow( hTask, SW_RESTORE );
		active = TRUE;
	}*/

	(this)->needRedraw = true;//CPU 20020501
}

void TaskbarButton::SelectDown(){
	container.SetCapture( this );

	mousePressed = TRUE;

	if(flashing)
		this->SetFlashing(FALSE);


/*CPU 20020501*/
	if ( active )
	{
		// already active, minimize it
		ShowWindowAsync( hTask, SW_MINIMIZE );
		active = FALSE;
	}
	else if (taskbar->LsBoxed && taskbar->lastActive == this->hTask && !IsIconic( this->hTask )){
		char tmpClass[BUFFER_SIZE];
		GetClassName(GetParent(taskbar->getHwnd()), tmpClass, 1024);
		if (strstr(tmpClass, "LsBoxClass")){
			// already active, minimize it
			ShowWindowAsync( hTask, SW_MINIMIZE );
			active = FALSE;
		}
	}
	else
	{
		if ( !IsIconic( hTask ))
		{
			// if not minimized, move to application's current desktop
			SendMessage( GetLitestepWnd(), LM_BRINGTOFRONT, 0, (LPARAM) hTask );
		}

		// undocumented call to switch window activation
		SwitchToThisWindow( hTask, SW_RESTORE );
		active = TRUE;
	}
/*end CPU 20020501*/

	container.Invalidate( x, y, width + 1, height + 1 );
}

void TaskbarButton::MenuUp(){
	mousePressed = FALSE;
	if ( container.GetCapture() != this )
		return ;

	container.ReleaseCapture();

	if ( !mouseOver )
	{
		container.Invalidate( x, y, width + 1, height + 1 );
		return ;
	}

	HMENU systemMenu;
	systemMenu = GetSystemMenu( hTask, FALSE );

	WINDOWPLACEMENT wp;
	wp.length = sizeof(WINDOWPLACEMENT);
	GetWindowPlacement( hTask, &wp );

	// restore is enabled only if minimized or maximized (not normal)
	EnableMenuItem( systemMenu, SC_RESTORE, MF_BYCOMMAND |
	                (wp.showCmd != SW_SHOWNORMAL ? MF_ENABLED : MF_GRAYED) );

	// move is enabled only if normal
	EnableMenuItem( systemMenu, SC_MOVE, MF_BYCOMMAND |
	                (wp.showCmd == SW_SHOWNORMAL ? MF_ENABLED : MF_GRAYED) );

	// size is enabled only if normal
	EnableMenuItem( systemMenu, SC_SIZE, MF_BYCOMMAND |
	                (wp.showCmd == SW_SHOWNORMAL ? MF_ENABLED : MF_GRAYED) );

	// minimize is enabled only if not minimized
	EnableMenuItem( systemMenu, SC_MOVE, MF_BYCOMMAND |
	                (wp.showCmd != SW_SHOWMINIMIZED ? MF_ENABLED : MF_GRAYED) );

	// maximize is enabled only if not maximized
	EnableMenuItem( systemMenu, SC_MOVE, MF_BYCOMMAND |
	                (wp.showCmd != SW_SHOWMAXIMIZED ? MF_ENABLED : MF_GRAYED) );


	// let application modify menu
	SendMessage( hTask, WM_INITMENUPOPUP, (WPARAM) systemMenu, MAKELPARAM( 0, TRUE ) );
	SendMessage( hTask, WM_INITMENU, (WPARAM) systemMenu, 0 );

	// display the menu
	POINT pt;
	GetCursorPos( &pt );

	int command = TrackPopupMenu( systemMenu,
	                              TPM_RETURNCMD | TPM_RIGHTBUTTON,
	                              pt.x, pt.y, 0,
	                              container.hWnd,
	                              NULL );

	if ( command )
	{
		if ( command != SC_MINIMIZE && command != SC_CLOSE )
			SetForegroundWindow( hTask );

		PostMessage( hTask, WM_SYSCOMMAND, (WPARAM) command, MAKELPARAM( pt.x, pt.y ) );
	}
	else
	{
		container.Invalidate( x, y, width + 1, height + 1 );
	}
}

void TaskbarButton::MenuDown(){
	container.SetCapture( this );

	mousePressed = TRUE;

	this->SetFlashing(FALSE);

	container.Invalidate( x, y, width + 1, height + 1 );
}

void TaskbarButton::OnMButtonDown( int, int )
{
	switch (taskbar->TaskBarTileMButton){
		case MOUSE_SELECT:
			SelectDown();
		    break;
		case MOUSE_MENU:
			MenuDown();
		    break;
		case MOUSE_ONTOP:
			OnTopToggle();
		    break;
		case MOUSE_CLOSE:
			Close();
		    break;
		case MOUSE_CLOSE2:
			Close2();
		    break;
		case MOUSE_NONE:
		    break;
        default:
		    break;
	}
}

void TaskbarButton::OnMButtonUp( int, int )
{
    
	switch (taskbar->TaskBarTileMButton){
		case MOUSE_SELECT:
			SelectUp();
		    break;
		case MOUSE_MENU:
			MenuUp();
		    break;
		case MOUSE_ONTOP:
			OnTopToggle();
		    break;
		case MOUSE_CLOSE:
			Close();
			break;
		case MOUSE_CLOSE2:
			Close2();
		    break;
		case MOUSE_NONE:
		    break;
        default:
		    break;
	}
}

void TaskbarButton::OnLButtonUp( int, int )
{
	switch (taskbar->TaskBarTileLButton){
		case MOUSE_SELECT:
			SelectUp();
		    break;
		case MOUSE_MENU:
			MenuUp();
		    break;
		case MOUSE_ONTOP:
			OnTopToggle();
			break;
		case MOUSE_CLOSE:
			Close();
		    break;
		case MOUSE_CLOSE2:
			Close2();
		    break;
		case MOUSE_NONE:
		    break;
        default:
		    break;
	}
}

void TaskbarButton::OnMouseMove( int mX, int mY )
{
	if ( mouseOver != HasPoint( mX, mY ) )
	{
		mouseOver = !mouseOver;
		container.Invalidate( x, y, width + 1, height + 1 );
	}
}

void TaskbarButton::OnPaint( HDC hDC )
{
	RECT r;

	r.left = x;
	r.top = y;
	r.right = x + width;
	r.bottom = y + height;

    TaskbarSkin *tempskin;

	int skinShift = 0;
	int textShift = 0;

    BOOL drawPressed;

      drawPressed = (GetForegroundWindow() == this->hTask) ;

	if( active ){
		taskbar->lastActive = this->hTask;
	}

	if(container.IsTaskFlash() && drawPressed){
		this->SetFlashing(FALSE);
		flashing = false;
		taskbar->flashCount = taskbar->flashCount - 1;
		//taskbar->flashCount = 0;
	}

    int state = 0;
	if(GetFlashing() && container.IsTaskFlash() && !IsOnTaskbarList(this->hTask, NoFlashList))
        state = 1;
	else if(IsIconic(this->hTask))
        state = 2;
	else if ( drawPressed )
        state = 3;
	else
		state = 0;

	/*char tmpTitle[BUFFER_SIZE], tmpClass[BUFFER_SIZE];
	GetWindowText(this->hTask, tmpTitle, 1024);
	GetClassName(this->hTask, tmpClass, 1024);
	LSLogPrintf(0,"Taskbar.dll"," %s - %s: %d",tmpClass, tmpTitle, state);*/

	if(state == 1)
		tempskin = GetFlashSkin();
	else if(state == 2)
		tempskin = GetMinimizedSkin();
	else if ( state == 3 )
	{
		//flash active
		tempskin = GetActiveSkin();

		skinShift = container.noSkinShift ? 0 : 1;
		textShift = container.noTextShift ? 0 : 1;
	}
	else{
		tempskin = GetSkin();
	}


	if ( !tempskin )
	{
		if ( !container.msTaskbar )
		{
			// simple 3D frame
			HBRUSH hbrBack;
			HPEN hpnFore;
			HPEN hpnTemp;
			COLORREF backcolor;
			if(state == 1)
				backcolor = container.flashColor;
			else if(state == 2)
				backcolor = container.minColor;
			else if ( state == 3 )
			{
				backcolor = container.activeColor;
			}
			else{
				backcolor = container.backColor;
			}

			hbrBack = CreateSolidBrush(backcolor);
			FillRect( hDC, &r, hbrBack );
			DeleteObject( hbrBack );

			hpnFore = CreatePen( PS_SOLID, 1, drawPressed ? container.foreColor1 : container.foreColor2 );
			hpnTemp = (HPEN) SelectObject( hDC, hpnFore );

			MoveToEx( hDC, r.left, r.bottom - 1, NULL );
			LineTo( hDC, r.right - 1, r.bottom - 1 );
			LineTo( hDC, r.right - 1, r.top );

			SelectObject( hDC, hpnTemp );
			DeleteObject( hpnFore );

			hpnFore = CreatePen( PS_SOLID, 1, drawPressed ? container.foreColor2 : container.foreColor1 );
			hpnTemp = (HPEN) SelectObject( hDC, hpnFore );

			LineTo( hDC, r.left, r.top );
			LineTo( hDC, r.left, r.bottom - 1 );

			SelectObject( hDC, hpnTemp );
			DeleteObject( hpnFore );
		}
		else
		{
			// MS taskbar style
			DrawEdge( hDC, &r, drawPressed ? EDGE_SUNKEN : EDGE_RAISED, BF_RECT | BF_MIDDLE | BF_SOFT );
		}
	}
	else
	{
		// simple 3D frame
		HBRUSH hbrBack;

		hbrBack = CreateSolidBrush( container.backColor );
		FillRect( hDC, &r, hbrBack );
		DeleteObject( hbrBack );

		// skin
		tempskin->Apply( hDC, x + skinShift, y + skinShift, width, height );
	}

	/*jesus666*/

	r.left += taskbar->buttonBorderLeft;
	r.top += taskbar->buttonBorderTop;
	r.right -= taskbar->buttonBorderRight;
	r.bottom -= taskbar->buttonBorderBottom;

	OffsetRect( &r, textShift, textShift );

	//if (taskbar->showIcon && icon && (r.right - r.left) >= taskbar->iconSize )//Jesus666
	if (taskbar->showIcon && icon)//Jesus666
	{
		int iconWidth = taskbar->iconSize, iconHeight =  taskbar->iconSize;
		int tmpIconX, tmpIconY;

		if(taskbar->iconX<0)
			tmpIconX = this->width / 2 + r.left - taskbar->iconSize / 2 - taskbar->buttonBorderLeft;
		else{
			tmpIconX = r.left + taskbar->iconX;
		}

		if(taskbar->iconY<0)
			tmpIconY = this->height / 2 + r.top - taskbar->iconSize / 2 - taskbar->buttonBorderTop;
		else{
			tmpIconY = r.top + taskbar->iconY;
		}

		if (((r.right - r.left) < taskbar->iconSize) || ((r.bottom - r.top) < taskbar->iconSize))
			iconHeight = iconWidth = min(r.right - r.left, r.bottom - r.top);

		DrawIconEx( hDC, tmpIconX, tmpIconY , icon, iconWidth, 
		            iconHeight, 0, NULL, DI_NORMAL );
	}

	if(taskbar->showText){//jesus666
		r.left += taskbar->buttonTextBorderLeft;
		r.top += taskbar->buttonTextBorderTop;
		r.right -= taskbar->buttonTextBorderRight;
		r.bottom -= taskbar->buttonTextBorderBottom;

		COLORREF tempColor;
		HFONT tempFont;

		if(state == 1)
		{
			tempColor = SetTextColor( hDC, container.textFlashColor );
			tempFont = (HFONT) SelectObject( hDC, container.flashFont );
			flashing = true;
			if( !IsOnTaskbarList(this->hTask, NoScrollList) )
				taskbar->flashCount = taskbar->flashCount + 1;
		}
		else if(state == 2)
		{
			tempColor = SetTextColor( hDC, container.textMinimizedColor );
			tempFont = (HFONT) SelectObject( hDC, container.minimizedFont );
		}
		else if ( state == 3 )
		{
	        tempColor = SetTextColor( hDC, container.textActiveColor );
			tempFont = (HFONT) SelectObject( hDC, container.activeFont );
		}
		else{
			tempColor = SetTextColor( hDC, container.textColor );
			tempFont = (HFONT) SelectObject( hDC, container.font );
		}

		SetBkMode( hDC, TRANSPARENT );
/*
uFormat 
[in] Unsigned integer that specifies the method of formatting the text. The following table shows the possible values that you can combine to create a value for this parameter. 
Value Description 
DT_BOTTOM Justifies the text to the bottom of the rectangle. You must combine this value with DT_SINGLELINE. 
DT_CALCRECT Determines the width and height of the rectangle. If the rectangle includes multiple lines of text, DrawText uses the width of the rectangle pointed to by the lpRect parameter and extends the base of the rectangle to bound the last line of text. If the rectangle includes only one line of text, DrawText modifies the right side of the rectangle so that it bounds the last character in the line. In either case, DrawText returns the height of the formatted text but does not draw the text. 
Before calling DrawText, an application must set the right and bottom members of the RECT structure pointed to by lpRect. These members are updated with the call to DrawText.
 
DT_CENTER Centers text horizontally in the rectangle. 
DT_EXPANDTABS Expands tab characters. The default number of characters per tab is eight.  
DT_INTERNAL Uses the system font to calculate text metrics.  
DT_LEFT Aligns text to the left. 
DT_NOCLIP Draws without clipping. DrawText is somewhat faster when DT_NOCLIP is used. 
DT_NOPREFIX Turns off processing of prefix characters. Normally, DrawText interprets the mnemonic-prefix character & as a directive to underscore the character that follows, and the mnemonic-prefix characters && as a directive to print a single &. By specifying DT_NOPREFIX, this processing is turned off. 
DT_RIGHT Aligns text to the right. 
DT_SINGLELINE Displays text on a single line only. Carriage returns and linefeeds do not break the line. 
DT_TABSTOP Sets tab stops. Bits 815, which form the high-order byte of the low-order word, of the uFormat parameter specify the number of characters for each tab. The default number of characters per tab is eight. You cannot use the DT_CALCRECT, DT_EXTERNALLEADING, DT_INTERNAL, DT_NOCLIP, and DT_NOPREFIX values with the DT_TABSTOP value. 
DT_TOP Top-justifies text. You must combine this value with DT_SINGLELINE. 
DT_VCENTER Centers text vertically. You must combine this value with DT_SINGLELINE. 
DT_WORDBREAK Breaks words. DrawText automatically breaks lines between words if a word would extend past the edge of the rectangle specified by the lpRect parameter. A carriage return-linefeed sequence also breaks the line. 
*/
		UINT position = DT_SINGLELINE | DT_END_ELLIPSIS; //jesus666
/*wtf with position >= 0?*/
		if(taskbar->textX == -1)
			position = position | DT_LEFT;
		else if(taskbar->textX == -2)
			position = position | DT_CENTER;
		else if(taskbar->textX == -3)
			position = position | DT_RIGHT;
	
		if(taskbar->textY == -1)
			position = position | DT_TOP;
		else if(taskbar->textY == -2)
			position = position | DT_VCENTER;
		else if(taskbar->textY == -3)
			position = position | DT_BOTTOM;

		if (hTask)
		{
			if (caption)
			{
				DrawText( hDC, caption, strlen( caption ), &r, position );
			}
		}
		else
		{
			if (otherCaption)
			{
				DrawText( hDC, otherCaption, strlen( otherCaption ), &r, position );
			}
		}

		SetTextColor( hDC, tempColor );
		SelectObject( hDC, tempFont );
	}
}

void TaskbarButton::OnRButtonDown( int, int )
{
	switch (taskbar->TaskBarTileRButton){
		case MOUSE_SELECT:
			SelectDown();
		    break;
		case MOUSE_MENU:
			MenuDown();
		    break;
		case MOUSE_ONTOP:
			OnTopToggle();
			break;
		case MOUSE_CLOSE:
			Close();
		    break;
		case MOUSE_CLOSE2:
			Close2();
		    break;
		case MOUSE_NONE:
		    break;
        default:
		    break;
	}
}

void TaskbarButton::OnRButtonUp( int, int )
{
	switch (taskbar->TaskBarTileRButton){
		case MOUSE_SELECT:
			SelectUp();
		    break;
		case MOUSE_MENU:
			MenuUp();
		    break;
		case MOUSE_ONTOP:
			OnTopToggle();
			break;
		case MOUSE_CLOSE:
			Close();
		    break;
		case MOUSE_CLOSE2:
			Close2();
		    break;
		case MOUSE_NONE:
		    break;
        default:
		    break;
	}
}

void TaskbarButton::OnUpdate(BOOL needRedraw)
{
	//it's not a force update, so let's check if we need to redraw
	if (!needRedraw) //CPU 20020501
	{
		// has activiation state changed?
		bool tempActive = (GetForegroundWindow() == hTask);
		//bool tempActive = (GetActiveWindow() == hTask);

		if ( tempActive != active )
		{
			active = tempActive;
			needRedraw = TRUE;
		}

		// has caption changed?
		LPSTR temp1, temp2;

		if (hTask)
		{
			temp1 = caption;
			temp2 = GetWindowCaption();
		}
		else
		{
			temp1 = otherCaption;
			temp2 = NULL;
		}

		if ( temp2 == NULL || strcmp( temp1, temp2 ) != 0 )
		{
			if (temp2)
			{
				delete [] caption;
				caption = temp2;
				needRedraw = TRUE;
				temp1 = caption;
			}

			// update tooltip
			TOOLINFO ti;
			memset( &ti, 0, sizeof(TOOLINFO) );

			ti.cbSize = sizeof(TOOLINFO);
			ti.hwnd = container.hWnd;
			ti.uId = (UINT) this;
			ti.lpszText = temp1;

			SendMessage( container.hToolTips, TTM_UPDATETIPTEXT, 0, (LPARAM) &ti );
		}
		else
		{
			if (temp2)
			{
				delete [] temp2;
			}
		}


		// has icon changed?
		HICON tempIcon = GetIcon();

		if ( icon != tempIcon )
		{
			icon = tempIcon;
			needRedraw = TRUE;
		}
	}

	if ( needRedraw ) //CPU 20020501
	  container.Invalidate( x, y, width + 1, height + 1 );
/*	if ( needRedraw ){
		taskbar->Layout();
	}*/
}

LPSTR TaskbarButton::GetWindowCaption()
{
	LPSTR windowCaption = NULL;
	int length;

	if (hTask)
	{
		length = GetWindowTextLength( hTask );

		windowCaption = new char[length + 1];
		GetWindowText( hTask, windowCaption, length + 1 );
	}

	return windowCaption;
}

HICON TaskbarButton::GetIcon()
{
	return GetIconFromWindow( hTask, (taskbar->iconSize>=32) ); //get a big icon sometimes
}

//
// TaskbarSkin
//

TaskbarSkin::TaskbarSkin()
{
	TaskbarSkin(NULL, NULL, NULL);
}

TaskbarSkin::TaskbarSkin( LPCSTR pszLeft, LPCSTR pszMiddle, LPCSTR pszRight )
{
	stretch = TRUE;
	backColor = 0;
	hbmLeft = NULL;
	hbmMiddle = NULL;
	hbmRight = NULL;

	if ( pszLeft && pszLeft[0] )
	{
		hbmLeft = LoadLSImage( pszLeft, NULL );
		GetLSBitmapSize( hbmLeft, &leftW, &leftH );
	}

	if ( pszMiddle && pszMiddle[0] )
	{
		hbmMiddle = LoadLSImage( pszMiddle, NULL );
		GetLSBitmapSize( hbmMiddle, &middleW, &middleH );
	}

	if ( pszRight && pszRight[0] )
	{
		hbmRight = LoadLSImage( pszRight, NULL );
		GetLSBitmapSize( hbmRight, &rightW, &rightH );
	}
}

TaskbarSkin::~TaskbarSkin()
{
	if ( hbmLeft )
		DeleteObject( hbmLeft );

	if ( hbmMiddle )
		DeleteObject( hbmMiddle );

	if ( hbmRight )
		DeleteObject( hbmRight );
}

void TaskbarSkin::Apply( HDC hDC, int x, int y, int width, int height )
{
	HDC hdcBuffer;
	HDC hdcMem;
	HBITMAP hbmBuffer;
	HGDIOBJ hgoBuffer;
	HGDIOBJ hgoMem;
	int widthLeft = width;

	hdcBuffer = CreateCompatibleDC( hDC );
	hbmBuffer = CreateCompatibleBitmap( hDC, width, height );
	hgoBuffer = SelectObject( hdcBuffer, hbmBuffer );

	hdcMem = CreateCompatibleDC( hDC );
	hgoMem;

	SetStretchBltMode( hdcBuffer, STRETCH_DELETESCANS );

	if (backColor)
	{
		// simple 3D frame
		RECT r;
		HBRUSH hbrBack;

		r.top = 0;
		r.left = 0;
		r.bottom = height;
		r.right = width;

		hbrBack = CreateSolidBrush( backColor );
		FillRect( hdcMem, &r, hbrBack );
		FillRect( hdcBuffer, &r, hbrBack );
		DeleteObject( hbrBack );
	}

	if ( hbmLeft && width > 0 )
	{
		hgoMem = SelectObject( hdcMem, hbmLeft );

		if (stretch)
		{
			StretchBlt( hdcBuffer, 0, 0, min( widthLeft, leftW ), height,
			            hdcMem, 0, 0, leftW, leftH, SRCCOPY );
		}
		else
		{
			BitBlt(hdcBuffer, 0, 0, min(widthLeft, leftW), height, hdcMem, 0, 0, SRCCOPY);
		}

		SelectObject( hdcMem, hgoMem );
		widthLeft = widthLeft - leftW;
	}

	if ( hbmRight && width > 0 )
	{
		hgoMem = SelectObject( hdcMem, hbmRight );

		if (stretch)
		{
			StretchBlt( hdcBuffer, width - rightW, 0, min( widthLeft, rightW ),
			            height, hdcMem, 0, 0, rightW, rightH, SRCCOPY );
		}
		else
		{
			BitBlt(hdcBuffer, width - rightW, 0, min(widthLeft, rightW), height, hdcMem, 0, 0, SRCCOPY);
		}

		SelectObject( hdcMem, hgoMem );
		widthLeft = widthLeft - rightW;
	}

	if ( hbmMiddle && width > 0 )
	{
		hgoMem = SelectObject( hdcMem, hbmMiddle );

		if (stretch)
		{
			StretchBlt( hdcBuffer, hbmLeft ? leftW : 0, 0, widthLeft, height,
			            hdcMem, 0, 0, middleW, middleH, SRCCOPY );
		}
		else
		{
			BitBlt(hdcBuffer, hbmLeft ? leftW : 0, 0, widthLeft, height, hdcMem, 0, 0, SRCCOPY);
		}

		SelectObject( hdcMem, hgoMem );
	}

	TransparentBltLS( hDC, x, y, width, height, hdcBuffer, 0, 0,
	                  RGB( 255, 0, 255 ) );

	SelectObject( hdcBuffer, hgoBuffer );
	DeleteObject( hbmBuffer );
	DeleteDC( hdcBuffer );

	DeleteDC( hdcMem );
}

void TaskbarButton::SetSkin(TaskbarSkin *s)
{
	skin = s;
}

void TaskbarButton::SetActiveSkin(TaskbarSkin *s)
{
	activeskin = s;
}

void TaskbarButton::SetFlashSkin(TaskbarSkin *s)
{
	flashskin = s;
}

void TaskbarButton::SetMinimizedSkin(TaskbarSkin *s)
{
	minimizedskin = s;
}

TaskbarSkin* TaskbarButton::GetSkin()
{
	return skin;
}

TaskbarSkin* TaskbarButton::GetActiveSkin()
{
	return activeskin;
}

TaskbarSkin* TaskbarButton::GetFlashSkin()
{
	return flashskin;
}


TaskbarSkin* TaskbarButton::GetMinimizedSkin()
{
	return minimizedskin;
}


void TaskbarSkin::SetStretch(BOOL s)
{
	stretch = s;
}

BOOL TaskbarSkin::GetStretch()
{
	return stretch;
}

void TaskbarSkin::SetBackColor(COLORREF color)
{
	backColor = color;
}

COLORREF TaskbarSkin::GetBackColor()
{
	return backColor;
}

void TaskbarButton::SetCaption(LPCSTR cap)
{
	if (!cap)
	{
		return ;
	}

	if (otherCaption)
	{
		delete [] otherCaption;
	}

	otherCaption = new char[strlen(cap) + 1];

	strcpy(otherCaption, cap);

	// add tooltip
	TOOLINFO ti;
	memset( &ti, 0, sizeof(TOOLINFO) );

	ti.cbSize = sizeof(TOOLINFO);
	ti.hwnd = container.hWnd;
	ti.uId = (UINT) this;
	ti.lpszText = otherCaption;

	SendMessage( container.hToolTips, TTM_ADDTOOL, 0, (LPARAM) &ti );
}
//to change
void TaskbarButton::Deactivate(){
	this->active = FALSE;
}

void BangBoxHook(HWND caller, const char *args)
{
	char *handle = strrchr(args,' ');

	if (handle) 
	{
		HWND boxwnd = (HWND)atoi(handle+1);
		if (boxwnd) 
		{
			HWND hWnd = taskbar->getHwnd();
			if (boxwnd != GetParent(hWnd))
			{
				SetWindowLong(hWnd, GWL_STYLE, (GetWindowLong(hWnd, GWL_STYLE) &~ WS_POPUP)|WS_CHILD);				
				SetParent(hWnd, boxwnd);
				RECT r;
				GetClientRect( boxwnd, &r );
				/*if (GetRCBoolDef("TaskbarPositioning", FALSE)){
					int x = taskbar->taskbarXS;
					int y = taskbar->taskbarYS;
					if(taskbar->taskbarXS<0)
						x = (r.right - r.left) + taskbar->taskbarXS;
					if(taskbar->taskbarYS<0)
						y = (r.bottom - r.top) + taskbar->taskbarYS;
					MoveWindow( taskbar->getHwnd(),  x, y, taskbar->width ,taskbar->height, TRUE);
				}*/
				if (taskbar->taskbarPositioning){//666  vrifier	
					int x = taskbar->taskbarXS;
					int y = taskbar->taskbarYS;
					if(taskbar->taskbarXS<0)
						x = (r.right - r.left) + taskbar->taskbarXS;
					if(taskbar->taskbarYS<0)
						y = (r.bottom - r.top) + taskbar->taskbarYS;
					MoveWindow( taskbar->getHwnd(),  x, y, taskbar->width ,taskbar->height, TRUE);
				}
				else{
					if ( taskbar->edge == EDGE_TOP )
					{
						MoveWindow( taskbar->getHwnd(),  0, 0, taskbar->width ,taskbar->height, TRUE);
					}
					else
					{
						MoveWindow( taskbar->getHwnd(),  0, r.bottom - r.top - taskbar->neededHeight, taskbar->width ,taskbar->height, TRUE);
					}
				}
		        showBangCommand(NULL, "");
				taskbar->LsBoxed = true;
			}
		}
	}
	return;
}

HWND Taskbar::getHwnd(){
	return hWnd;
}